<!--
 Licensed to the Apache Software Foundation (ASF) under one
 or more contributor license agreements.  See the NOTICE file
 distributed with this work for additional information
 regarding copyright ownership.  The ASF licenses this file
 to you under the Apache License, Version 2.0 (the
 "License"); you may not use this file except in compliance
 with the License.  You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
-->


<head>
  <title>Getting started</title>
</head>

### Clone and build the Twill library

```sh
$ git clone https://git-wip-us.apache.org/repos/asf/incubator-twill.git twill
$ cd twill
$ mvn clean install
```

### Quick example

Let's begin by building a basic `EchoServer` in Twill. Traditionally, when you build a server as simple as this,
you add logic within a `Runnable` implementation to run it in a `Thread` using an appropriate `ExecutorService`:

```java
public class EchoServer implements Runnable {
  private static Logger LOG = LoggerFactory.getLogger(EchoServer.class);
  private final ServerSocket serverSocket;

  public EchoServer() {
    ...
  }

  @Override
  public void run() {
    while ( isRunning() ) {
      Socket socket = serverSocket.accept();
      ...
    }
  }
}
```

Our example defines an implementation of `Runnable` that implements the `run()` method.
The `EchoServer` is now a `Runnable` that can be executed by an `ExecutorService`:

```java
ExecutorService service = Executors.newFixedThreadPool(2);
service.submit(new EchoServer());
```

The above model is familiar, but now assume you want to run your `EchoServer` on a YARN cluster.
To do this, all you need to do is implement the `TwillRunnable` interface similarly to how you
normally implement `Runnable`:

```java
public class EchoServer implements TwillRunnable {

  private static Logger LOG = LoggerFactory.getLogger(EchoServer.class);
  private final ServerSocket serverSocket;
  private final int port;

  public EchoServer() {
    ...
  }

  @Override
  public void run() {
    while ( isRunning() ) {
      Socket socket = serverSocket.accept();
      ...
    }
  }
}
```

To run `EchoServer` on the YARN cluster, you need a `TwillRunnerService`, which is similar to `ExecutorService`.
You can specify the YARN cluster configuration and connection string to a running instance of a Zookeeper service
to create an instance of `TwillRunnerService`:

```java
TwillRunnerService runnerService = new YarnTwillRunnerService(
  new YarnConfiguration(), zkConnectStr);
runnerService.startAndWait();
```

Now you are read to run the EchoServer on YARN, by simply preparing and starting through the `TwillRunnerService`. You
can also attach log handler to receives logs generated by EchoService running on some nodes in the cluster:

```java
TwillController controller = runnerService.prepare(new EchoServer())
  .addLogHandler(new PrinterLogHandler(new PrintWriter(System.out)))
  .start();
```

Now that the `EchoServer` is started and you can inspect or control the application through the `TwillController`.
For example, you can attach listeners for state changes:

```java
controller.addListener(new ListenerAdapter() {
  @Override
  public void running() {
    LOG.info('Echo Server Started');
  }
}
```

You can also stop the application through the `TwillController`:

```java
controller.stop().get();
```

This will shut down the application master and all the containers.